import pandas as pd

from xl2times import utils


class TestUtils:
    def test_explode(self):
        """Test that explode logics functions correctly."""
        input_df1 = pd.DataFrame(
            [
                ("PRC1", 100, pd.NA, 200),
                ("PRC2", 100, 150, pd.NA),
                ("PRC3", pd.NA, 150, 200),
            ],
            columns=["process", "act_bnd", "actbnd", "act_bnd"],
        )

        input_df2 = pd.DataFrame(
            [
                ("PRC1", 100, pd.NA),
                ("PRC2", 100, 150),
                ("PRC3", pd.NA, 150),
            ],
            columns=["process", "act_bnd", "actbnd"],
        )

        data_cols1 = ["act_bnd", "actbnd", "act_bnd"]
        data_cols2 = ["act_bnd", "actbnd"]

        correct_index1 = pd.RangeIndex(9)
        correct_index2 = pd.RangeIndex(6)

        correct_result1 = (
            pd.DataFrame(
                [
                    ("PRC1", 100),
                    ("PRC1", pd.NA),
                    ("PRC1", 200),
                    ("PRC2", 100),
                    ("PRC2", 150),
                    ("PRC2", pd.NA),
                    ("PRC3", pd.NA),
                    ("PRC3", 150),
                    ("PRC3", 200),
                ],
                columns=["process", "value"],
                index=correct_index1,
                dtype=object,
            ),
            pd.Series(
                [
                    "act_bnd",
                    "actbnd",
                    "act_bnd",
                    "act_bnd",
                    "actbnd",
                    "act_bnd",
                    "act_bnd",
                    "actbnd",
                    "act_bnd",
                ],
                index=correct_index1,
            ),
        )

        correct_result2 = (
            pd.DataFrame(
                [
                    ("PRC1", 100),
                    ("PRC1", pd.NA),
                    ("PRC2", 100),
                    ("PRC2", 150),
                    ("PRC3", pd.NA),
                    ("PRC3", 150),
                ],
                columns=["process", "value"],
                index=correct_index2,
                dtype=object,
            ),
            pd.Series(
                ["act_bnd", "actbnd", "act_bnd", "actbnd", "act_bnd", "actbnd"],
                index=correct_index2,
            ),
        )

        output1 = utils.explode(input_df1, data_cols1)
        output2 = utils.explode(input_df2, data_cols2)

        assert output1[0].equals(correct_result1[0]), "Dataframes should be equal"
        assert output1[1].equals(correct_result1[1]), "Series should be equal"
        assert output2[0].equals(correct_result2[0]), "Dataframes should be equal"
        assert output2[1].equals(correct_result2[1]), "Series should be equal"


if __name__ == "__main__":
    TestUtils().test_explode()



import argparse
import hashlib
import os
import pickle
import sys
import time
from collections import defaultdict
from concurrent.futures import ProcessPoolExecutor
from datetime import datetime, timedelta
from io import StringIO
from pathlib import Path

import numpy as np
import pandas as pd
from loguru import logger
from pandas.core.frame import DataFrame

from . import excel, transforms, utils
from .datatypes import Config, DataModule, EmbeddedXlTable, TimesModel

_log_sep = "=" * 80 + "\n"


cache_dir = Path.home() / ".cache/xl2times/"
cache_dir.mkdir(exist_ok=True, parents=True)


def invalidate_cache(max_age: timedelta = timedelta(days=365)):
    """
    Delete any cache files older than max_age.

    Args:
        max_age: Maximum age of a cache file to be considered valid. Any cache files older than this are deleted.
    """
    for file in cache_dir.glob("*.pkl"):
        if datetime.now() - datetime.fromtimestamp(file.lstat().st_mtime) > max_age:
            try:
                file.unlink()
            except Exception as e:
                logger.warning(f"Failed to delete old cache file {file}. {e}")


def _read_xlsx_cached(filename: str | Path) -> list[EmbeddedXlTable]:
    """Extract EmbeddedXlTables from xlsx file (cached).

    Since excel.extract_tables is quite slow, we cache its results in `cache_dir`.
    Each cache file is named {filename}_{hash}.pkl, and contains a pickled
    `[EmbeddedXlTable]`.

    Args:
        filename: Path to the xlsx file to extract tables from.
    """
    filename = Path(filename).resolve()
    with filename.open("rb") as f:
        digest = hashlib.file_digest(f, "sha256")  # pyright: ignore
    hsh = digest.hexdigest()
    cached_file = (cache_dir / f"{Path(filename).stem}_{hsh}.pkl").resolve()

    if cached_file.exists():
        # just load and return the cached pickle
        with cached_file.open("rb") as f:
            tables = pickle.load(f)
            logger.info(f"Using cached data for {filename} from {cached_file}")
    else:
        # extract data and write it to cache before returning it
        tables = excel.extract_tables(str(filename))
        with cached_file.open("wb") as f:
            pickle.dump(tables, f)
        logger.info(f"Saved cache for {filename} to {cached_file}")

    return tables


def convert_xl_to_times(
    input_files: list[str],
    output_dir: str,
    config: Config,
    model: TimesModel,
    no_cache: bool,
    stop_after_read: bool = False,
) -> dict[str, DataFrame]:
    start_time = datetime.now()

    invalidate_cache()
    with ProcessPoolExecutor(utils.max_workers) as executor:
        raw_tables = executor.map(
            excel.extract_tables if no_cache else _read_xlsx_cached, input_files
        )
    # raw_tables is a list of lists, so flatten it:
    raw_tables = [t for ts in raw_tables for t in ts]
    logger.info(
        f"Extracted (potentially cached) {len(raw_tables)} tables,"
        f" {sum(table.dataframe.shape[0] for table in raw_tables)} rows"
        f" in {datetime.now() - start_time}"
    )

    if stop_after_read:
        # Convert absolute paths to relative paths to enable comparing raw_tables.txt across machines
        raw_tables.sort(key=lambda x: (x.filename, x.sheetname, x.range))
        input_dir = os.path.commonpath([t.filename for t in raw_tables])
        raw_tables = [strip_filename_prefix(t, input_dir) for t in raw_tables]

    dump_tables(raw_tables, os.path.join(output_dir, "raw_tables.txt"))
    if stop_after_read:
        return {}

    transform_list = [
        transforms.normalize_tags_columns,
        transforms.remove_fill_tables,
        lambda config, tables, model: [
            transforms.remove_comment_cols(t) for t in tables
        ],
        transforms.validate_input_tables,
        transforms.remove_tables_with_formulas,  # slow
        transforms.normalize_column_aliases,
        transforms.remove_comment_rows,
        transforms.revalidate_input_tables,
        transforms.capitalise_table_values,
        transforms.process_regions,
        transforms.convert_com_tables,
        transforms.process_time_periods,
        transforms.remove_exreg_cols,
        transforms.generate_dummy_processes,
        transforms.process_time_slices,
        transforms.process_transform_table_variants,
        transforms.apply_tag_specified_defaults,
        transforms.process_transform_tables,
        transforms.process_transform_availability,
        transforms.process_flexible_import_tables,  # slow
        transforms.process_user_constraint_tables,
        transforms.harmonise_tradelinks,
        transforms.include_tables_source,
        transforms.process_processes,
        transforms.create_model_topology,
        transforms.fill_in_column_defaults,
        transforms.generate_uc_properties,
        transforms.expand_rows_parallel,  # slow
        transforms.process_tradelinks,
        transforms.merge_tables,
        transforms.remove_invalid_values,
        transforms.include_cgs_in_topology,
        transforms.fill_in_missing_pcgs,
        transforms.complete_processes,
        transforms.create_model_units,
        transforms.process_wildcards,
        transforms.convert_aliases,
        transforms.enforce_availability,
        transforms.complete_model_trade,
        transforms.create_model_cgs,
        transforms.prepare_for_querying,
        transforms.apply_transform_tables,
        transforms.generate_implied_topology,
        transforms.verify_uc_topology,
        transforms.explode_process_commodity_cols,
        transforms.apply_final_fixup,
        transforms.assign_model_attributes,
        transforms.resolve_remaining_cgs,
        lambda config, tables, model: dump_tables(
            tables, os.path.join(output_dir, "merged_tables.txt")
        ),
        transforms.complete_dictionary,
        transforms.convert_to_string,
        lambda config, tables, model: produce_times_tables(config, tables, model),
    ]

    input = raw_tables
    output = {}
    for transform in transform_list:
        start_time = time.time()
        output = transform(config, input, model)
        end_time = time.time()
        logger.opt(raw=True).debug(_log_sep)
        logger.info(
            f"transform {transform.__code__.co_name} took {end_time - start_time:.2f} seconds"
        )
        logger.opt(raw=True).debug(_log_sep)
        # Way to conditionally evaluate the table dump only on debug log level
        # https://loguru.readthedocs.io/en/stable/overview.html#lazy-evaluation-of-expensive-functions
        logger.opt(lazy=True).debug(
            "All tables:\n{dump}", dump=lambda: _all_table_dump(output)
        )
        input = output
    assert isinstance(output, dict)

    logger.info(
        f"Conversion complete, {len(output)} tables produced,"
        f" {sum(df.shape[0] for df in output.values())} rows"
    )

    return output


def _all_table_dump(tables: list[EmbeddedXlTable] | dict[str, DataFrame]) -> str:
    """A dump of current values in all tables, for debugging."""
    result = StringIO()
    if isinstance(tables, list):
        for table in sorted(
            tables, key=lambda t: (t.tag, t.filename, t.sheetname, t.range)
        ):
            result.write(str(table))
            result.write("\n")
    elif isinstance(tables, dict):
        for tag, df in tables.items():
            df_str = df.to_csv(index=False, lineterminator="\n")
            result.write(f"{tag}\n{df_str}{df.shape}\n")
    return result.getvalue()


def write_csv_tables(tables: dict[str, DataFrame], output_dir: str):
    os.makedirs(output_dir, exist_ok=True)
    for item in os.listdir(output_dir):
        if item.endswith(".csv"):
            os.remove(os.path.join(output_dir, item))
    for tablename, df in tables.items():
        df.to_csv(os.path.join(output_dir, tablename + "_output.csv"), index=False)
    logger.success(
        f"Excel files successfully converted to CSV and written to {output_dir}"
    )


def read_csv_tables(input_dir: str) -> dict[str, DataFrame]:
    result = {}
    csv_files = list(Path(input_dir).glob("*.csv"))
    for filename in csv_files:
        result[filename.stem] = pd.read_csv(filename)
    return result


def compare(
    data: dict[str, DataFrame], ground_truth: dict[str, DataFrame], output_dir: str
) -> str:
    logger.info(
        f"Ground truth contains {len(ground_truth)} tables,"
        f" {sum(df.shape[0] for _, df in ground_truth.items())} rows"
    )

    missing = set(ground_truth.keys()).difference(data.keys())
    missing_str = ", ".join(
        [f"{x} ({ground_truth[x].shape[0]})" for x in sorted(missing)]
    )
    if len(missing) > 0:
        logger.info(f"Missing {len(missing)} tables: {missing_str}")

    additional_tables = set(data.keys()).difference(ground_truth.keys())
    additional_str = ", ".join(
        [f"{x} ({data[x].shape[0]})" for x in sorted(additional_tables)]
    )
    if len(additional_tables) > 0:
        logger.info(f"{len(additional_tables)} additional tables: {additional_str}")
    # Additional rows starts as the sum of lengths of additional tables produced
    total_additional_rows = sum(len(data[x]) for x in additional_tables)

    total_gt_rows = 0
    total_correct_rows = 0
    for table_name, gt_table in sorted(
        ground_truth.items(), reverse=True, key=lambda t: len(t[1])
    ):
        if table_name in data:
            data_table = data[table_name]

            # Remove .integer suffix added to duplicate column names by CSV reader (mangle_dupe_cols=False not supported)
            transformed_gt_cols = [col.split(".")[0] for col in gt_table.columns]
            data_cols = list(data_table.columns)
            if transformed_gt_cols != data_cols:
                logger.info(
                    f"Table {table_name} header incorrect, was"
                    f" {data_cols}, should be {transformed_gt_cols}"
                )

            # Convert rows to lowercase strings for case-insensitive comparison
            gt_rows = set(str(row).lower() for row in gt_table.to_numpy().tolist())
            data_rows = set(str(row).lower() for row in data_table.to_numpy().tolist())
            total_gt_rows += len(gt_rows)
            total_correct_rows += len(gt_rows.intersection(data_rows))
            additional = data_rows.difference(gt_rows)
            total_additional_rows += len(additional)
            missing = gt_rows.difference(data_rows)
            if len(additional) != 0 or len(missing) != 0:
                logger.info(
                    f"Table {table_name} ({data_table.shape[0]} rows,"
                    f" {gt_table.shape[0]} GT rows) contains {len(additional)}"
                    f" additional rows and is missing {len(missing)} rows"
                )
            if len(additional) != 0:
                DataFrame(additional).to_csv(
                    os.path.join(output_dir, table_name + "_additional.csv"),
                    index=False,
                )
            if len(missing) != 0:
                DataFrame(missing).to_csv(
                    os.path.join(output_dir, table_name + "_missing.csv"),
                    index=False,
                )
    result = (
        f"{(total_correct_rows / total_gt_rows) if total_gt_rows!=0 else np.nan :.1%} of ground truth rows present"
        f" in output ({total_correct_rows}/{total_gt_rows})"
        f", {total_additional_rows} additional rows"
    )

    logger.success(result)
    return result


def produce_times_tables(
    config: Config, input: dict[str, DataFrame], model: TimesModel
) -> dict[str, DataFrame]:
    logger.info(
        f"produce_times_tables: {len(input)} tables incoming,"
        f" {sum(len(value) for (_, value) in input.items())} rows"
    )
    file_order = defaultdict(lambda: -1)
    for i, f in enumerate(model.files):
        file_order[f] = i
    # Keep only those mappings for which parameters that are defined in the input
    par_tables = {"Attributes", "UCAttributes"}.intersection(input.keys())
    defined_pars = set()
    for table in par_tables:
        defined_pars = defined_pars.union(set(input[table]["attribute"]))
    mappings = [
        m
        for m in config.times_xl_maps
        if m.xl_name not in par_tables or m.filter_rows.get("attribute") in defined_pars
    ]

    def keep_last_by_file_order(df):
        """Drop duplicate rows, keeping the last dupicate row (including value) as per
        input file order, and remove the `source_filename` column from the DataFrame.

        Note: we do not remove duplicate values for the same query columns for parameters
        here, because in the future we might want to re-use the processed tables and
        select the rows coming from different scenarios/files after processing just once.
        If so, at that point we can use the info in the `source_filename` column to do
        this.
        """
        if "source_filename" in df.columns:
            df["file_order"] = df["source_filename"].map(file_order)
            df = df.sort_values(by="file_order", kind="stable")
            df = df.drop(columns=["source_filename", "file_order"])
        df = df.drop_duplicates(keep="last")
        return df.reset_index(drop=True)

    result = {}
    used_tables = set()
    for mapping in mappings:
        if mapping.xl_name not in input:
            logger.info(
                f"Cannot produce table {mapping.times_name} because"
                f" {mapping.xl_name} does not exist"
            )
        else:
            used_tables.add(mapping.xl_name)
            df = input[mapping.xl_name]
            # Filter rows according to filter_rows mapping:
            for filter_col, filter_val in mapping.filter_rows.items():
                if filter_col not in df.columns:
                    logger.info(
                        f"Cannot produce table {mapping.times_name} because"
                        f" {mapping.xl_name} does not contain column {filter_col}"
                    )
                    # TODO break this loop and continue outer loop?
                filter = set(x.lower() for x in (filter_val,))
                i = df[filter_col].str.lower().isin(filter)
                df = df[i]
            if not set(mapping.xl_cols).issubset(df.columns):
                missing = set(mapping.xl_cols).difference(df.columns)
                logger.info(
                    f"Cannot produce table {mapping.times_name} because"
                    f" {mapping.xl_name} does not contain the required columns"
                    f" - {', '.join(missing)}"
                )
            else:
                # Ensure that df is not a view
                df = df.reset_index(drop=True)
                # Excel columns can be duplicated into multiple Times columns
                for times_col, xl_col in mapping.col_map.items():
                    df[times_col] = df[xl_col]
                # Keep only the required columns
                cols_to_keep = set(mapping.times_cols).union({"source_filename"})
                cols_to_drop = [x for x in df.columns if x not in cols_to_keep]
                df = df.drop(columns=cols_to_drop)
                # Drop duplicates, keeping last seen rows as per file order
                df = keep_last_by_file_order(df)
                # Drop rows with missing values
                # TODO this is a hack. Use pd.StringDtype() so that notna() is sufficient
                i = (
                    df[mapping.times_cols[-1]].notna()
                    & (df != "None").all(axis=1)
                    & (df != "nan").all(axis=1)
                    & (df != "").all(axis=1)
                    & (df != "<NA>").all(axis=1)
                )
                df = df.loc[i, mapping.times_cols]
                # Drop tables that are empty after filtering and dropping Nones:
                if len(df) == 0:
                    continue
                result[mapping.times_name] = df

    unused_tables = set(input.keys()).difference(used_tables)
    if len(unused_tables) > 0:
        logger.info(
            f"{len(unused_tables)} unused tables: {', '.join(sorted(unused_tables))}"
        )

    return result


def write_dd_files(tables: dict[str, DataFrame], config: Config, output_dir: str):
    encoding = "utf-8"
    os.makedirs(output_dir, exist_ok=True)
    for item in os.listdir(output_dir):
        if item.endswith(".dd"):
            os.remove(os.path.join(output_dir, item))

    def convert_set(df: DataFrame):
        has_description = "TEXT" in df.columns
        # Remove duplicate rows, ignoring text column
        if has_description:
            query_columns = [c for c in df.columns if c != "TEXT"]
            df = df.drop_duplicates(subset=query_columns, keep="last")
        for row in df.itertuples(index=False):
            row_str = "'.'".join(
                (str(x) for k, x in row._asdict().items() if k != "TEXT")
            )
            desc = f" '{row.TEXT}'" if has_description else ""
            yield f"'{row_str}'{desc}\n"

    def convert_parameter(tablename: str, df: DataFrame):
        if "VALUE" not in df.columns:
            raise KeyError(f"Unable to find VALUE column in parameter {tablename}")
        # Remove duplicate rows, ignoring value column
        query_columns = [c for c in df.columns if c != "VALUE"] or None
        df = df.drop_duplicates(subset=query_columns, keep="last")
        for row in df.itertuples(index=False):
            val = row.VALUE
            row_str = "'.'".join(
                (str(x) for k, x in row._asdict().items() if k != "VALUE")
            )
            yield f"'{row_str}' {val}\n" if row_str else f"{val}\n"

    sets = {m.times_name for m in config.times_xl_maps if "VALUE" not in m.col_map}

    # Compute map fname -> tables: put ALL_TS and MILESTONYR in separate files
    tables_in_file = {
        "ts.dd": ["ALL_TS"],
        "milestonyr.dd": ["MILESTONYR"],
        "output.dd": [
            t for t in config.dd_table_order if t not in ["ALL_TS", "MILESTONYR"]
        ],
    }

    for fname, tablenames in tables_in_file.items():
        with open(os.path.join(output_dir, fname), "w", encoding=encoding) as fout:
            for tablename in [t for t in tablenames if t in tables]:
                df = tables[tablename]
                if tablename in sets:
                    fout.write(f"SET {tablename}\n/\n")
                    lines = convert_set(df)
                else:
                    fout.write(f"PARAMETER\n{tablename} ' '/\n")
                    lines = convert_parameter(tablename, df)
                # Sort lines to ensure consistent output, except for ALL_TS
                if tablename != "ALL_TS":
                    lines = sorted(lines)
                fout.writelines(lines)
                fout.write("\n/;\n")

    logger.success(
        f"Excel files successfully converted to DD and written to {output_dir}"
    )


def strip_filename_prefix(table, prefix):
    if isinstance(table, EmbeddedXlTable):
        if table.filename.startswith(prefix):
            table.filename = table.filename[len(prefix) + 1 :]
    return table


def dump_tables(tables: list, filename: str) -> list:
    os.makedirs(os.path.dirname(filename), exist_ok=True)
    with open(filename, "w", encoding="utf-8") as text_file:
        for t in tables if isinstance(tables, list) else tables.items():
            if isinstance(t, EmbeddedXlTable):
                tag = t.tag
                text_file.write(f"sheetname: {t.sheetname}\n")
                text_file.write(f"range: {t.range}\n")
                text_file.write(f"filename: {t.filename}\n")
                if t.uc_sets:
                    text_file.write(f"uc_sets: {t.uc_sets}\n")
                df = t.dataframe
            else:
                tag = t[0]
                df = t[1]
            text_file.write(f"tag: {tag}\n")
            types = ", ".join([f"{i} ({v})" for i, v in df.dtypes.items()])
            text_file.write(f"types: {types}\n")
            text_file.write(df.to_csv(index=False, lineterminator="\n"))
            text_file.write("\n" * 2)

    return tables


def run(args: argparse.Namespace) -> str | None:
    """Runs the xl2times conversion.

    Parameters
    ----------
    args
        Pre-parsed command line arguments

    Returns
    -------
        comparison with ground-truth string if `ground_truth_dir` is provided, else None.
    """
    utils.setup_logger(args.verbose)

    config = Config(
        "times_mapping.txt",
        "times-info.json",
        "times-sets.json",
        "veda-tags.json",
        "veda-attr-defaults.json",
        args.regions,
        args.include_dummy_imports,
    )

    model = TimesModel()

    if not isinstance(args.input, list) or len(args.input) < 1:
        logger.critical(f"expected at least 1 input. Got {args.input}")
        sys.exit(-1)
    elif len(args.input) == 1:
        assert os.path.isdir(args.input[0])
        input_files = [
            str(path)
            for path in Path(args.input[0]).rglob("*")
            if path.suffix in [".xlsx", ".xlsm"] and not path.name.startswith("~")
        ]
        if utils.is_veda_based(input_files):
            input_files = utils.filter_veda_filename_patterns(input_files)
        logger.info(f"Loading {len(input_files)} files from {args.input[0]}")
    else:
        input_files = args.input

    model.files = [Path(path).stem for path in input_files]

    processing_order = ["base", "subres", "trade", "demand", "scen", "syssettings"]
    for data_module in processing_order:
        model.data_modules = model.data_modules + sorted(
            [
                item
                for item in {
                    DataModule.module_name(path)
                    for path in input_files
                    if DataModule.module_type(path) == data_module
                }
                if item is not None
            ]
        )

    if args.only_read:
        tables = convert_xl_to_times(
            input_files,
            args.output_dir,
            config,
            model,
            args.no_cache,
            stop_after_read=True,
        )
        sys.exit(0)

    tables = convert_xl_to_times(
        input_files, args.output_dir, config, model, args.no_cache
    )

    if args.dd:
        write_dd_files(tables, config, args.output_dir)
    else:
        write_csv_tables(tables, args.output_dir)

    if args.ground_truth_dir:
        ground_truth = read_csv_tables(args.ground_truth_dir)
        # Use the same convert_to_string transform on GT so that comparisons are fair
        ground_truth = transforms.convert_to_string(config, ground_truth, model)
        comparison = compare(tables, ground_truth, args.output_dir)
        return comparison
    else:
        return None


def parse_args(arg_list: None | list[str]) -> argparse.Namespace:
    """Parses command line arguments.

    Parameters
    ----------
    arg_list
        List of command line arguments. Uses sys.argv (default argparse behaviour) if `None`.

    Returns
    -------
    argparse.Namespace
        Parsed arguments.
    """
    args_parser = argparse.ArgumentParser()
    args_parser.add_argument(
        "input",
        nargs="*",
        help="Either an input directory, or a list of input xlsx/xlsm files to process",
    )
    args_parser.add_argument(
        "--regions",
        type=str,
        default="",
        help="Comma-separated list of regions to include in the model",
    )
    args_parser.add_argument(
        "--include_dummy_imports",
        action="store_true",
        help="Include dummy import processes in the model",
    )
    args_parser.add_argument(
        "--output_dir", type=str, default="output", help="Output directory"
    )
    args_parser.add_argument(
        "--ground_truth_dir",
        type=str,
        help="Ground truth directory to compare with output",
    )
    args_parser.add_argument("--dd", action="store_true", help="Output DD files")
    args_parser.add_argument(
        "--only_read",
        action="store_true",
        help="Read xlsx/xlsm files and stop after outputting raw_tables.txt",
    )
    args_parser.add_argument(
        "--no_cache",
        action="store_true",
        help="Ignore cache and re-extract tables from XLSX files",
    )
    args_parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        help="Verbosity. Multiple `-v`s increase the log level. Can also be set on the command line by setting the environment variable `LOGURU_LEVEL`. Available levels are `TRACE`, `DEBUG`, `INFO`, `SUCCESS`, `WARNING`, `ERROR`, and `CRITICAL`. Default is `SUCCESS`",
    )
    args = args_parser.parse_args(arg_list)
    return args


def main(arg_list: None | list[str] = None) -> None:
    """Main entry point for the xl2times package.

    Returns
    -------
        None.
    """
    args = parse_args(arg_list)
    run(args)


if __name__ == "__main__":
    main(sys.argv[1:])
    sys.exit(0)



from __future__ import (
    annotations,
)

# see https://loguru.readthedocs.io/en/stable/api/type_hints.html#module-autodoc_stub_file.loguru
import functools
import gzip
import os
import pickle
import sys
from dataclasses import replace
from math import floor, log10
from pathlib import Path, PurePath

import numpy
import pandas as pd
from loguru import logger
from more_itertools import one
from pandas.core.frame import DataFrame

from . import datatypes

# prevent excessive number of processes in Windows and high cpu-count machines
# TODO make this a cli param or global setting?
max_workers: int = 4 if os.name == "nt" else min(16, os.cpu_count() or 16)


def apply_composite_tag(table: datatypes.EmbeddedXlTable) -> datatypes.EmbeddedXlTable:
    """Handles table level declarations.

    Declarations can be included in the table tag and will apply to all data that doesn't
    have a different value for that index specified. For example, ~FI_T: DEMAND would
    assign DEMAND as the attribute for all values in the table that don't have an
    attribute specification at the column or row level. After applying the declaration
    this function will return the modified table with the simplified table tag (e.g.
    ~FI_T).

    See page 15 of https://iea-etsap.org/docs/Documentation_for_the_TIMES_Model-Part-IV.pdf
    for more context.

    Parameters
    ----------
    table
        Table in EmbeddedXlTable format.

    Returns
    -------
    datatypes.EmbeddedXlTable
        Table in EmbeddedXlTable format with declarations applied
        and table tag simplified.
    """
    defaults = table.defaults
    if not defaults:
        return table

    df = table.dataframe
    # Check for ANSWER-style defaults
    if "=" in defaults:
        # Split multiple comma-separated defaults / make defaults a list
        defaults = defaults.split(",")
        # Check whether there are invalid values on the list
        invalid_defaults = [default for default in defaults if "=" not in default]
        if invalid_defaults:
            logger.warning(f"Expected ANSWER-style defaults, got {invalid_defaults}")
        defaults = [default.split("=") for default in defaults if "=" in default]
        # TODO: check whether a column is allowed in a particular table type
        for col, val in defaults:
            colname = col.lower()
            if colname in df.columns:
                df[colname] = df[colname].fillna(val.upper())
            else:
                df[colname] = val.upper()
    else:
        # TODO: Resolve the default value (it doesn't have to be an attribute)
        if "attribute" not in df.columns:
            df["attribute"] = pd.NA
        df["attribute"] = df["attribute"].fillna(defaults.upper())
    return replace(table, dataframe=df)


def explode(df: DataFrame, data_columns: list[str]) -> tuple[DataFrame, pd.Series]:
    """Transpose the 'data_columns' in each row into a column of values, replicating the
    other columns. The name for the new column is "VALUE".

    Parameters
    ----------
    df
        Dataframe to be exploded.
    data_columns
        Names of the columns to be exploded.

    Returns
    -------
    tuple[DataFrame,pd.Series]
        Tuple with the exploded dataframe and a Series of the original
        column name for each value in each new row.
    """
    # Handle duplicate columns (https://pandas.pydata.org/docs/user_guide/duplicates.html)
    if len(set(data_columns)) < len(data_columns):
        cols = df.columns.to_list()
        data_cols_idx = [idx for idx, val in enumerate(cols) if val in data_columns]
        data = df.iloc[:, data_cols_idx].values.tolist()
    else:
        data = df[data_columns].values.tolist()

    other_columns = [
        colname for colname in df.columns.values if colname not in data_columns
    ]
    df = df[other_columns]
    value_column = "value"
    df = df.assign(value=data)
    nrows = df.shape[0]
    df = df.explode(value_column, ignore_index=True)
    names = pd.Series(data_columns * nrows, index=df.index, dtype=str)

    return df, names


def single_table(
    tables: list[datatypes.EmbeddedXlTable], tag: str
) -> datatypes.EmbeddedXlTable:
    """Make sure exactly one table in 'tables' has the given table tag, and return it.
    If there are none or more than one raise an error.

    Parameters
    ----------
    tables
        List of tables in EmbeddedXlTable format.
    tag
        Tag name.

    Returns
    -------
    datatypes.EmbeddedXlTable
        Table with the given tag in EmbeddedXlTable format.
    """
    return one(table for table in tables if table.tag == tag)


def single_column(
    tables: list[datatypes.EmbeddedXlTable], tag: str, colname: str
) -> numpy.ndarray:
    """Make sure exactly one table in 'tables' has the given table tag, and return the
    values for the given column name. If there are none or more than one raise an error.

    Parameters
    ----------
    tables
        List of tables in EmbeddedXlTable format.
    tag
        Tag name.
    colname
        Column name to return the values of.

    Returns
    -------
    numpy.ndarray
        Values for the column in the given table.
    """
    return single_table(tables, tag).dataframe[colname].values


def merge_columns(
    tables: list[datatypes.EmbeddedXlTable], tag: str, colname: str
) -> numpy.ndarray:
    """Return a list with all the values belonging to a column 'colname' from a table
    with the given tag.

    Parameters
    ----------
    tables
        List of tables in EmbeddedXlTable format.
    tag
        Tag name to select tables
    colname
        Column name to select values.

    Returns
    -------
    numpy.ndarray
        List of values for the given column name and tag.
    """
    columns = [table.dataframe[colname].values for table in tables if table.tag == tag]
    return numpy.concatenate(columns)


def missing_value_inherit(df: DataFrame, colname: str) -> None:
    """For each None value in the specifed column of the dataframe, replace it with the
    last non-None value. If no previous non-None value is found leave it as it is. This
    function modifies the supplied dataframe and returns None.

    Parameters
    ----------
    df
        Dataframe to be filled in.
    colname
        Name of the column to be filled in.

    Returns
    -------
    None
        None. The dataframe is filled in in place.
    """
    # TODO: should we use pandas.DataFrame.fillna(method="ffill") instead?
    last = None
    for index, value in df[colname].items():
        if value is None:
            df.loc[index, colname] = last
        else:
            last = value


def get_scalar(table_tag: str, tables: list[datatypes.EmbeddedXlTable]):
    table = one(filter(lambda t: t.tag == table_tag, tables))
    if table.dataframe.shape[0] != 1 or table.dataframe.shape[1] != 1:
        raise ValueError("Not scalar table")
    return table.dataframe["value"].values[0]


def has_negative_patterns(pattern: str) -> bool:
    if len(pattern) == 0:
        return False
    return pattern[0] == "-" or ",-" in pattern


def remove_negative_patterns(pattern: str) -> str:
    # Remove trailing commas
    pattern = pattern.rstrip(",")
    if len(pattern) == 0:
        return pattern
    return ",".join([word for word in pattern.split(",") if word[0] != "-"])


def remove_positive_patterns(pattern: str) -> str:
    # Remove trailing commas
    pattern = pattern.rstrip(",")
    if len(pattern) == 0:
        return pattern
    return ",".join([word[1:] for word in pattern.split(",") if word[0] == "-"])


def remove_whitespace(pattern: str) -> str:
    return ",".join([word.strip() for word in pattern.split(",")])


@functools.lru_cache(maxsize=int(1e6))
def create_regexp(pattern: str, combined: bool = True) -> str:
    pattern = remove_whitespace(pattern)
    # Exclude negative patterns
    if has_negative_patterns(pattern):
        pattern = remove_negative_patterns(pattern)
    # Handle comma-separated values
    pattern = pattern.replace(",", r"$|^")
    if len(pattern) == 0:
        return r".*"  # matches everything
    # Substite VEDA wildcards with regex patterns; escape metacharacters.
    # ("_", ".") and ("[.]", "_") are meant to apply one after another to handle
    # the usage of "_" equivalent to "?" and "[_]" as literal "_".
    substitute = [(".", "\\."), ("_", "."), ("[.]", "_"), ("*", ".*"), ("?", ".")]
    for old, new in substitute:
        pattern = pattern.replace(old, new)
    # Do not match substrings
    pattern = rf"^{pattern}$"
    return pattern


@functools.lru_cache(maxsize=int(1e6))
def create_negative_regexp(pattern: str) -> str:
    pattern = remove_whitespace(pattern)
    # Exclude positive patterns
    pattern = remove_positive_patterns(pattern)
    if len(pattern) == 0:
        pattern = r"^$"  # matches nothing
    return create_regexp(pattern)


def round_sig(x, sig_figs):
    if x == 0.0:
        return 0.0
    return round(x, -int(floor(log10(abs(x)))) + sig_figs - 1)


# Get entry point file name as default log name
default_log_name = Path(sys.argv[0]).stem
default_log_name = "log" if default_log_name == "" else default_log_name


def _case_insensitive_match(path: str, pattern: str) -> bool:
    """Do case-insensitive path match. Convert to lowercase first, because
    case_sensitive parameter in match is not available before Python 3.12.
    """
    return PurePath(path.lower()).match(pattern.lower())


def is_veda_based(files: list[str]) -> bool:
    """Determine whether the model follows Veda file structure.
    This function does not verify file extensions.
    """
    marker = "SysSettings.*"

    matches = [file for file in files if _case_insensitive_match(file, marker)]

    if len(matches) == 1:
        return True
    elif len(matches) > 1:
        raise ValueError(f"Only one {marker} expected. Multiple detected: {matches}")
    else:
        return False


def filter_veda_filename_patterns(files: list[str]) -> list[str]:
    """Filter files by patterns recognised by Veda.
    This function does not verify file extensions.
    """
    legal_paths = (
        "BY_Trans.*",
        "LMA*.*",
        "Set*.*",
        "SysSettings.*",
        "VT_*.*",
        "SubRES_TMPL/SubRES_*.*",
        "SuppXLS/Demands/Dem_Alloc+Series.*",
        "SuppXLS/Demands/ScenDem_*.*",
        "SuppXLS/ParScenFiles/Scen_Par-*.*",
        "SuppXLS/Scen_*.*",
        "SuppXLS/Trades/ScenTrade_*.*",
    )
    # Generate a set of fiels that match the patterns
    filtered = {
        file
        for file in files
        for legal_path in legal_paths
        if _case_insensitive_match(file, legal_path)
    }
    # Return as a list
    return list(filtered)


def set_log_level(level: int | None) -> str:
    """Sets the log level, in order of priority, to the provided int `level`, the
    `LOGURU_LEVEL` environment variable, or `WARNING` by default.

    E.g. `os.environ["LOGURU_LEVEL"] = "INFO"`
    Available levels are `TRACE`, `DEBUG`, `INFO`, `SUCCESS`, `WARNING`, `ERROR`, and
    `CRITICAL`. Default is `SUCCESS` which is level `0`, and higher levels are more
    verbose.
    """
    level_map = {
        3: "TRACE",
        2: "DEBUG",
        1: "INFO",
        0: "SUCCESS",
        -1: "WARNING",
        -2: "ERROR",
        -3: "CRITICAL",
    }
    # First priority is argument `level`
    if level is not None:
        return level_map[level]
    # Second, if env var is set, let's roll with that
    env_level = os.getenv("LOGURU_LEVEL")
    if env_level is not None:
        return env_level
    # Default log level
    return level_map[0]


def setup_logger(
    level: int | None, log_name: str = default_log_name, log_dir: str = "."
):
    """Configure loguru.

    Call this once from entrypoints to set up a new logger.
    In non-entrypoint modules, just use `from loguru import logger` directly.

    Log file will be written to `f"{log_dir}/{log_name}.log"`

    Parameters
    ----------
    log_name
        Name of the log. Corresponding log file will be called {log_name}.log. (Default value = default_log_name)
    log_dir
        Directory to write the log file to. Default is the current working directory.
    """
    log_level = set_log_level(level)

    base_format = "<cyan>{time:YYYY-MM-DD HH:mm:ss.SSS}</cyan> | <level>{level: >8}</level> : <level>{message}</level>"
    filename_and_thread = '(<cyan>{name}:{thread.name}:pid-{process}</cyan> "<cyan>{file.path}</cyan>:<cyan>{line}</cyan>")'
    if level is not None and level > 1:
        stdout_format = base_format + filename_and_thread
    else:
        stdout_format = base_format

    logger.remove()
    logger.add(
        sink=sys.stdout,
        diagnose=True,
        level=log_level,
        format=stdout_format,
    )
    logger.add(
        sink=f"{log_dir}/{log_name}.log",
        enqueue=True,
        mode="a+",
        level="INFO",
        format=base_format + filename_and_thread,
        colorize=False,
        serialize=False,
        diagnose=False,
        rotation="20 MB",
        compression="zip",
    )


def save_state(
    config: datatypes.Config,
    tables: dict[str, DataFrame],
    model: datatypes.TimesModel,
    filename: str,
) -> None:
    """Save the state from a transform step to a single pickle file.

    Useful for troubleshooting regressions by diffing with state from another branch.
    """
    pickle.dump({"tables": tables, "model": model}, gzip.open(filename, "wb"))
    logger.debug(f"State saved to {filename}")


def compare_df_dict(
    df_before: dict[str, DataFrame],
    df_after: dict[str, DataFrame],
    sort_cols: bool = True,
    context_rows: int = 2,
) -> None:
    """Simple function to compare two dictionaries of DataFrames.

    Parameters
    ----------
    df_before
        the first dictionary of DataFrames to compare
    df_after
        the second dictionary of DataFrames to compare
    sort_cols
        whether to sort the columns before comparing.  Set True if the column order
        is unimportant. (Default value = True)
    context_rows
        number of rows to show around the first difference (Default value = 2)
    """
    for key in df_before:
        before = df_before[key]
        after = df_after[key]

        if sort_cols:
            before = before.sort_index(axis="columns")
            after = after.sort_index(axis="columns")

        if not before.equals(after):
            # print first line that is different, and its surrounding lines
            for i in range(len(before)):
                if not before.columns.equals(after.columns):
                    logger.warning(
                        f"Table {key} has different columns (or column order):\n"
                        f"BEFORE: {before.columns}\n"
                        f"AFTER: {after.columns}"
                    )
                    break
                if not before.iloc[i].equals(after.iloc[i]):
                    logger.warning(
                        f"Table {key} is different, first difference at row {i}:\n"
                        f"BEFORE:\n{before.iloc[i - context_rows:i + context_rows + 1]}\n"
                        f"AFTER: \n{after.iloc[i - context_rows:i + context_rows + 1]}"
                    )
                    break
        else:
            logger.success(f"Table {key} is the same")


def diff_state(
    filename_before: str, filename_after: str, sort_cols: bool = False
) -> None:
    """Diffs dataframes from two persisted state files created with save_state().

    Typical usage:
    - Save the state from a branch with a regression at some point in the transforms:
    - Switch to `main` branch and save the state from the same point:
    - Diff the two states:

    For example:

    >>> from utils import save_state, diff_state
    >>> save_state(config, tables, model, "branch.pkl.gz")
    >>> save_state(config, tables, model, "main.pkl.gz")
    >>> diff_state("branch.pkl.gz", "main.pkl.gz")
    """
    # TODO also compare config and non-dataframe model attributes?
    before = pickle.load(gzip.open(filename_before, "rb"))
    after = pickle.load(gzip.open(filename_after, "rb"))

    # Compare DFs in the tables dict
    logger.info("Comparing `table` dataframes...")
    compare_df_dict(before["tables"], after["tables"], sort_cols=sort_cols)

    # Compare DFs on the model object
    model_before = before["model"]
    model_after = after["model"]
    dfs_before = {
        a: getattr(model_before, a)
        for a in dir(model_before)
        if isinstance(getattr(model_before, a), pd.DataFrame)
    }
    dfs_after = {
        a: getattr(model_after, a)
        for a in dir(model_after)
        if isinstance(getattr(model_after, a), pd.DataFrame)
    }
    logger.info("Comparing `model` dataframes...")
    compare_df_dict(dfs_before, dfs_after, sort_cols=sort_cols)



import json
import re
from collections import defaultdict
from collections.abc import Iterable
from dataclasses import dataclass, field
from enum import Enum
from functools import cached_property
from importlib import resources
from itertools import chain
from pathlib import PurePath

from loguru import logger
from pandas.core.frame import DataFrame

# ============================================================================
# ===============                   CLASSES                   ================
# ============================================================================


class Tag(str, Enum):
    """Enum class to enumerate all the accepted table tags by this program.

    You can see a list of all the possible tags in section 2.4 of
    https://iea-etsap.org/docs/Documentation_for_the_TIMES_Model-Part-IV.pdf
    """

    active_p_def = "~ACTIVEPDEF"
    book_regions_map = "~BOOKREGIONS_MAP"
    comagg = "~COMAGG"
    comemi = "~COMEMI"
    currencies = "~CURRENCIES"
    defaultyear = "~DEFAULTYEAR"
    def_units = "~DEFUNITS"
    endyear = "~ENDYEAR"
    fi_comm = "~FI_COMM"
    fi_process = "~FI_PROCESS"
    fi_t = "~FI_T"
    milestoneyears = "~MILESTONEYEARS"
    start_year = "~STARTYEAR"
    tfm_ava = "~TFM_AVA"
    tfm_ava_c = "~TFM_AVA-C"
    tfm_comgrp = "~TFM_COMGRP"
    tfm_csets = "~TFM_CSETS"
    tfm_dins = "~TFM_DINS"
    tfm_dins_at = "~TFM_DINS-AT"
    tfm_dins_ts = "~TFM_DINS-TS"
    tfm_dins_tsl = "~TFM_DINS-TSL"
    tfm_fill = "~TFM_FILL"
    tfm_fill_r = "~TFM_FILL-R"
    tfm_ins = "~TFM_INS"
    tfm_ins_at = "~TFM_INS-AT"
    tfm_ins_ts = "~TFM_INS-TS"
    tfm_ins_tsl = "~TFM_INS-TSL"
    tfm_ins_txt = "~TFM_INS-TXT"
    tfm_mig = "~TFM_MIG"
    tfm_psets = "~TFM_PSETS"
    tfm_topdins = "~TFM_TOPDINS"
    tfm_topins = "~TFM_TOPINS"
    tfm_upd = "~TFM_UPD"
    tfm_upd_at = "~TFM_UPD-AT"
    tfm_upd_ts = "~TFM_UPD-TS"
    time_periods = "~TIMEPERIODS"
    time_slices = "~TIMESLICES"
    tradelinks = "~TRADELINKS"
    tradelinks_dins = "~TRADELINKS_DINS"
    uc_sets = "~UC_SETS"
    uc_t = "~UC_T"
    # This is used by Veda for unit conversion when displaying results
    unitconversion = "~UNITCONVERSION"

    @classmethod
    def has_tag(cls, tag: str) -> bool:
        return tag in cls._value2member_map_


class DataModule(str, Enum):
    """Categorise data into modules based on the file they are coming from."""

    base = "VT_*.*, BY_Trans.*"
    syssettings = "SysSettings.*"
    subres = "SubRES_TMPL/SubRES_*.*"
    sets = "Set*.*"
    lma = "LMA*.*"
    demand = "SuppXLS/Demands/Dem_Alloc+Series.*, SuppXLS/Demands/ScenDem_*.*"
    scen = "SuppXLS/Scen_*.*"
    trade = "SuppXLS/Trades/ScenTrade_*.*"

    @classmethod
    def determine_type(cls, path: str) -> "DataModule | None":
        for data_module in cls:
            if any(
                PurePath(path.lower()).match(pattern.lower().strip())
                for pattern in data_module.value.split(",")
            ):
                return data_module
        return None

    @classmethod
    def module_type(cls, path: str) -> str | None:
        module_type = cls.determine_type(path)
        if module_type:
            return module_type.name
        else:
            return None

    @classmethod
    def submodule(cls, path: str) -> str | None:
        match cls.determine_type(path):
            case DataModule.base | DataModule.subres:
                if PurePath(path.lower()).match("*_trans.*"):
                    return "trans"
                else:
                    return "main"
            case DataModule.trade:
                if PurePath(path.lower()).match("scentrade__trade_links.*"):
                    return "main"
                else:
                    return "trans"
            case DataModule.demand:
                if PurePath(path.lower()).match("dem_alloc+series.*"):
                    return "main"
                else:
                    return "trans"
            case None:
                return None
            case _:
                return "main"

    @classmethod
    def module_name(cls, path: str) -> str | None:
        module_type = cls.determine_type(path)
        match module_type:
            case (
                DataModule.base
                | DataModule.sets
                | DataModule.lma
                | DataModule.demand
                | DataModule.trade
                | DataModule.syssettings
            ):
                return module_type.name.upper()
            case DataModule.subres:
                return re.sub(
                    "^SUBRES_", "", re.sub("_TRANS$", "", PurePath(path).stem.upper())
                )
            case DataModule.scen:
                return re.sub("^SCEN_", "", PurePath(path).stem.upper())
            case None:
                return None


@dataclass
class EmbeddedXlTable:
    """A table object: a pandas dataframe wrapped with some metadata.

    Attributes
    ----------
    tag
        Table tag associated with this table in the excel file used as input.
    defaults
        Defaults for the table that are separated by a colon from the tag.
    uc_sets
        User constrained tables are declared with tags which indicate their type and domain of coverage.
    sheetname
        Name of the excel worksheet where this table was extracted from.
    range
        Range of rows and columns that contained this table in the original excel worksheet.
    filename
        Name of the original excel file where this table was extracted from.
    dataframe
        Pandas dataframe containing the values of the table.
    """

    tag: str
    uc_sets: dict[str, str]
    sheetname: str
    range: str
    filename: str
    dataframe: DataFrame
    defaults: str | None = field(default=None)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, EmbeddedXlTable):
            return False
        return (
            self.tag == o.tag
            and self.uc_sets == o.uc_sets
            and self.range == o.range
            and self.filename == o.filename
            and self.dataframe.shape == o.dataframe.shape
            and (
                len(self.dataframe) == 0  # Empty tables don't affect our output
                or self.dataframe.sort_index(axis=1).equals(
                    o.dataframe.sort_index(axis=1)
                )
            )
        )

    def __str__(self) -> str:
        df_str = self.dataframe.to_csv(index=False, lineterminator="\n")
        return f"EmbeddedXlTable(tag={self.tag}, uc_sets={self.uc_sets}, sheetname={self.sheetname}, range={self.range}, filename={self.filename}, dataframe=\n{df_str}{self.dataframe.shape})"


@dataclass
class TimesXlMap:
    """The mapping between the TIMES excel tables used by the tool for input and the
    transformed tables it outputs. The mappings are defined in the
    `times_mapping.txt` and `times-info.json` files.

    Attributes
    ----------
    times_name
        Name of the table in its output form.
    times_cols
        Name of the columns that the table will have in its output form. They will be
        in the header of the output csv files.
    xl_name
        Tag for the Excel table used as input. You can see a list of all the possible
        tags in section 2.4 of
        https://iea-etsap.org/docs/Documentation_for_the_TIMES_Model-Part-IV.pdf
    xl_cols
        Columns from the Excel table used as input.
    col_map
        A mapping from Excel column names to Times column names.
    filter_rows
        A map from column name to value to filter rows to. If `{}`, all rows are
        outputted. E.g., `{'Attribute': 'COM_ELAST'}`
    """

    times_name: str
    times_cols: list[str]
    xl_name: str  # TODO once we move away from times_mapping.txt, make this type Tag
    xl_cols: list[str]
    col_map: dict[str, str]
    filter_rows: dict[str, str]


@dataclass
class TimesModel:
    """A class containing all the information about the processed TIMES model."""

    internal_regions: set[str] = field(default_factory=set)
    all_regions: set[str] = field(default_factory=set)
    processes: DataFrame = field(default_factory=DataFrame)
    commodities: DataFrame = field(default_factory=DataFrame)
    commodity_groups: DataFrame = field(default_factory=DataFrame)
    topology: DataFrame = field(default_factory=DataFrame)
    implied_topology: DataFrame = field(default_factory=DataFrame)
    trade: DataFrame = field(default_factory=DataFrame)
    attributes: DataFrame = field(default_factory=DataFrame)
    user_constraints: DataFrame = field(default_factory=DataFrame)
    uc_attributes: DataFrame = field(default_factory=DataFrame)
    ts_tslvl: DataFrame = field(default_factory=DataFrame)
    ts_map: DataFrame = field(default_factory=DataFrame)
    time_periods: DataFrame = field(default_factory=DataFrame)
    units: DataFrame = field(default_factory=DataFrame)
    start_year: int = field(default_factory=int)
    files: list[str] = field(default_factory=list)
    data_modules: list[str] = field(default_factory=list)
    custom_sets: DataFrame = field(default_factory=DataFrame)

    @property
    def external_regions(self) -> set[str]:
        return self.all_regions.difference(self.internal_regions)

    @property
    def data_years(self) -> set[int]:
        """data_years are years for which there is data specified."""
        data_years = set()
        for attributes in [self.attributes, self.uc_attributes]:
            if not attributes.empty:
                # Index of the year column with non-empty values
                # index = attributes["year"] != ""
                # data_years.update(attributes["year"][index].astype(int).values)
                # TODO: Ensure that non-parseble vals don't get this far
                int_years = attributes["year"].astype(
                    int, errors="ignore"
                )  # leave non-parseable vals alone
                int_years = [
                    y for y in int_years if isinstance(y, int)
                ]  # remove non-parseable years
                data_years.update(int_years)
        # Remove interpolation rules before return
        return {y for y in data_years if y >= 1000}

    @property
    def past_years(self) -> set[int]:
        """Pastyears is the set of all years before start_year."""
        return {x for x in self.data_years if x < self.start_year}

    @property
    def model_years(self) -> set[int]:
        """model_years is the union of past_years and the representative years of the
        model (middleyears).
        """
        return self.past_years | set(self.time_periods["m"].values)

    # TODO: Invalidate and recompute the below property when self.topology changes.
    @cached_property
    def veda_cgs(self) -> dict[tuple[str, str, str], str]:
        """A dictionary mapping commodities to their Veda commodity groups."""
        cols = ["region", "process", "commodity", "csets"]
        # Exclude auxillary flows
        index = self.topology["io"].isin({"IN", "OUT"})
        veda_cgs = self.topology[cols + ["io"]][index]
        veda_cgs = veda_cgs.drop_duplicates(subset=cols, keep="last")
        veda_cgs["veda_cg"] = veda_cgs["csets"] + veda_cgs["io"].str[:1]
        veda_cgs = veda_cgs.set_index(["region", "process", "commodity"])[
            "veda_cg"
        ].to_dict()
        return veda_cgs


class Config:
    """Encapsulates all configuration options for a run of the tool, including the
    mapping betwen excel tables and output tables, categories of tables, etc.
    """

    times_xl_maps: list[TimesXlMap]
    dd_table_order: Iterable[str]
    all_attributes: set[str]
    attr_aliases: set[str]
    # Attribute by type (commodity, process, flow)
    attr_by_type: dict[str, set[str]]
    # Attribute by index (region, year, etc.)
    attr_by_index: dict[str, set[str]]
    # For each tag, this dictionary maps each column alias to the normalized name
    column_aliases: dict[Tag, dict[str, str]]
    # For each tag, this dictionary maps each column name to its default value
    column_default_value: dict[Tag, dict[str, str]]
    # For each tag, this dictionary specifies comment row symbols by column name
    row_comment_chars: dict[Tag, dict[str, list]]
    # List of tags for which empty tables should be discarded
    discard_if_empty: Iterable[Tag]
    veda_attr_defaults: dict[str, dict[str, list]]
    # Known columns for each tag
    known_columns: dict[Tag, set[str]]
    # Query columns for each tag
    query_columns: dict[Tag, set[str]]
    # Columns that may contain lists for each tag
    lists_columns: dict[Tag, set[str]]
    # Required columns for each tag
    required_columns: dict[Tag, set[str]]
    # Names of regions to include in the model; if empty, all regions are included.
    filter_regions: set[str]
    times_sets: dict[str, list[str]]
    # Switch to prevent overwriting of I/E settings in BASE and SubRES
    ie_override_in_syssettings: bool = False
    # Switch to include dummy imports in the model
    include_dummy_imports: bool

    def __init__(
        self,
        mapping_file: str,
        times_info_file: str,
        times_sets_file: str,
        veda_tags_file: str,
        veda_attr_defaults_file: str,
        regions: str,
        include_dummy_imports: bool,
    ):
        self.times_xl_maps = Config._read_mappings(mapping_file)
        (
            self.dd_table_order,
            self.all_attributes,
            self.attr_by_type,
            self.attr_by_index,
            param_mappings,
        ) = Config._process_times_info(times_info_file)
        self.times_sets = Config._read_times_sets(times_sets_file)
        (
            self.column_aliases,
            self.column_default_value,
            self.row_comment_chars,
            self.discard_if_empty,
            self.query_columns,
            self.lists_columns,
            self.known_columns,
            self.required_columns,
            self.add_columns,
            self.forward_fill_cols,
        ) = Config._read_veda_tags_info(veda_tags_file)
        self.veda_attr_defaults, self.attr_aliases = Config._read_veda_attr_defaults(
            veda_attr_defaults_file, param_mappings
        )
        # Migration in progress: use parameter mappings from times_info_file for now
        name_to_map = {m.times_name: m for m in self.times_xl_maps}
        for m in param_mappings:
            name_to_map[m.times_name] = m
        self.times_xl_maps = list(name_to_map.values())
        self.filter_regions = Config._read_regions_filter(regions)
        self.include_dummy_imports = include_dummy_imports

    @staticmethod
    def _read_times_sets(
        times_sets_file: str,
    ) -> dict[str, list[str]]:
        # Read times_sets_file
        with resources.open_text("xl2times.config", times_sets_file) as f:
            times_sets = json.load(f)

        return times_sets

    @staticmethod
    def _process_times_info(
        times_info_file: str,
    ) -> tuple[
        Iterable[str],
        set[str],
        dict[str, set[str]],
        dict[str, set[str]],
        list[TimesXlMap],
    ]:
        # Read times_info_file and compute dd_table_order:
        # We output tables in order by categories: set, subset, subsubset, md-set, and parameter
        with resources.open_text("xl2times.config", times_info_file) as f:
            table_info = json.load(f)
        categories = ["set", "subset", "subsubset", "md-set", "parameter"]
        cat_to_tables = defaultdict(list)
        for item in table_info:
            cat_to_tables[item["gams-cat"]].append(item["name"])
        unknown_cats = {item["gams-cat"] for item in table_info} - set(categories)
        if unknown_cats:
            logger.warning(f"Unknown categories in times-info.json: {unknown_cats}")
        dd_table_order = chain.from_iterable(
            sorted(cat_to_tables[c]) for c in categories
        )

        # Compute the set of all attributes, i.e. all entities with category = parameter
        attributes = {
            item["name"].upper()
            for item in table_info
            if item["gams-cat"] == "parameter"
        }

        # Determine the attributes by type
        attr_by_type = dict()

        attr_type_conditions = {
            "commodity": {"commodity": True, "process": False},
            "process": {"process": True, "commodity": False},
            "flow": {"process": True, "commodity": True},
        }

        for attr_type, conditions in attr_type_conditions.items():
            attr_by_type[attr_type] = {
                attr["name"]
                for attr in table_info
                if all(
                    (index in attr["mapping"]) is is_present
                    for index, is_present in conditions.items()
                )
            }

        # Categorise attributes by index (region, year, etc.)
        attr_by_index = dict()
        # Set of all indexes in the mapping
        all_indexes = set()
        for attr in table_info:
            if attr["gams-cat"] == "parameter":
                all_indexes = all_indexes.union(attr["mapping"])
        for index in all_indexes:
            list_of_attrs = set()
            for attr in table_info:
                if attr["gams-cat"] != "parameter":
                    continue
                if index in attr["mapping"]:
                    list_of_attrs.add(attr["name"])
            attr_by_index[index] = list_of_attrs

        # Compute the mapping for attributes / parameters:
        def create_mapping(entity):
            assert entity["gams-cat"] == "parameter"
            times_cols = entity["indexes"] + ["VALUE"]
            xl_cols = entity["mapping"] + ["value"]  # TODO map in json
            col_map = dict(zip(times_cols, xl_cols))
            # If tag starts with UC, then the data is in UCAttributes, else Attributes
            xl_name = (
                "UCAttributes"
                if entity["name"].lower().startswith("uc")
                else "Attributes"
            )
            return TimesXlMap(
                times_name=entity["name"],
                times_cols=times_cols,
                xl_name=xl_name,
                xl_cols=xl_cols,
                col_map=col_map,
                filter_rows={"attribute": entity["name"]},  # TODO value:1?
            )

        param_mappings = [
            create_mapping(x)
            for x in table_info
            if x["gams-cat"] == "parameter"
            and "type" not in x  # TODO Generalise derived parameters?
        ]

        return dd_table_order, attributes, attr_by_type, attr_by_index, param_mappings

    @staticmethod
    def _read_mappings(filename: str) -> list[TimesXlMap]:
        """Function to load mappings from a text file between the excel sheets we use as
        input and the tables we give as output.

        The mappings have the following structure:

        OUTPUT_TABLE[DATAYEAR,VALUE] = ~TimePeriods(Year,B)

        where OUTPUT_TABLE is the name of the table we output and it includes a list of the
        different fields or column names it includes. On the other side, TimePeriods is the type
        of table that we will use as input to produce that table, and the arguments are the
        columns of that table to use to produce the output. The last argument can be of the
        form `Attribute:ATTRNAME` which means the output will be filtered to only the rows of
        the input table that have `ATTRNAME` in the Attribute column.

        The mappings are loaded into TimesXlMap objects. See the description of that class for more
        information of the different fields they contain.

        Parameters
        ----------
        filename
            Name of the text file containing the mappings.

        Returns
        -------
        list[TimesXlMap]
            List of mappings in TimesXlMap format.
        """
        mappings = []
        dropped = []
        with resources.open_text("xl2times.config", filename) as file:
            while True:
                line = file.readline().rstrip()
                if line == "":
                    break
                (times, xl) = line.split(" = ")
                (times_name, times_cols_str) = list(
                    filter(None, re.split(r"\[|\]", times))
                )
                (xl_name, xl_cols_str) = list(filter(None, re.split(r"\(|\)", xl)))
                times_cols = times_cols_str.split(",")
                xl_cols = xl_cols_str.split(",")
                filter_rows = {}
                for i, s in enumerate(xl_cols):
                    if ":" in s:
                        [col_name, col_val] = s.split(":")
                        filter_rows[col_name.strip().lower()] = col_val.strip()
                xl_cols = [s.lower() for s in xl_cols if ":" not in s]

                # TODO remove: Filter out mappings that are not yet finished
                if xl_name != "~TODO" and not any(
                    c.startswith("TODO") for c in xl_cols
                ):
                    col_map = {}
                    assert len(times_cols) <= len(xl_cols)
                    for index, value in enumerate(times_cols):
                        col_map[value] = xl_cols[index]
                    # Uppercase and validate tags:
                    if xl_name.startswith("~"):
                        xl_name = xl_name.upper()
                    entry = TimesXlMap(
                        times_name=times_name,
                        times_cols=times_cols,
                        xl_name=xl_name,
                        xl_cols=xl_cols,
                        col_map=col_map,
                        filter_rows=filter_rows,
                    )
                    mappings.append(entry)
                else:
                    dropped.append(line)

        if len(dropped) > 0:
            logger.info(f"Dropping {len(dropped)} mappings that are not yet complete")
        return mappings

    @staticmethod
    def _read_veda_tags_info(
        veda_tags_file: str,
    ) -> tuple[
        dict[Tag, dict[str, str]],
        dict[Tag, dict[str, str]],
        dict[Tag, dict[str, list]],
        Iterable[Tag],
        dict[Tag, set[str]],
        dict[Tag, set[str]],
        dict[Tag, set[str]],
        dict[Tag, set[str]],
        dict[Tag, set[str]],
        dict[Tag, set[str]],
    ]:
        def to_tag(s: str) -> Tag:
            # The file stores the tag name in lowercase, and without the ~
            return Tag("~" + s.upper())

        # Read veda_tags_file
        with resources.open_text("xl2times.config", veda_tags_file) as f:
            veda_tags_info = json.load(f)

        # Check that all the tags we use are present in veda_tags_file
        tags = {to_tag(tag_info["tag_name"]) for tag_info in veda_tags_info}
        for tag in Tag:
            if tag not in tags:
                logger.info(
                    f"WARNING: datatypes.Tag has an unknown Tag {tag} not in {veda_tags_file}"
                )

        valid_column_names = {}
        column_default_value = {}
        row_comment_chars = {}
        discard_if_empty = []
        query_cols = defaultdict(set)
        lists_cols = defaultdict(set)
        known_cols = defaultdict(set)
        required_cols = defaultdict(set)
        add_cols = defaultdict(set)
        forward_fill_cols = defaultdict(set)

        for tag_info in veda_tags_info:
            tag_name = to_tag(tag_info["tag_name"])
            if "valid_fields" in tag_info:
                discard_if_empty.append(tag_name)
                valid_column_names[tag_name] = {}
                column_default_value[tag_name] = {}
                row_comment_chars[tag_name] = {}
                # Process column aliases and comment chars:
                for valid_field in tag_info["valid_fields"]:
                    valid_field_names = valid_field.get("aliases", list())
                    if (
                        "use_name" in valid_field
                        and valid_field["use_name"] != valid_field["name"]
                    ):
                        field_name = valid_field["use_name"]
                        valid_field_names.append(valid_field["name"])
                    else:
                        field_name = valid_field["name"]

                    if "default_to" in valid_field:
                        column_default_value[tag_name][field_name] = valid_field[
                            "default_to"
                        ]

                    if valid_field.get("query_field", False):
                        query_cols[tag_name].add(field_name)

                    if valid_field.get("comma-separated-list", False):
                        lists_cols[tag_name].add(field_name)

                    if valid_field.get("add_if_absent", False):
                        add_cols[tag_name].add(field_name)

                    if valid_field.get("remove_any_row_if_absent", False):
                        required_cols[tag_name].add(field_name)

                    if valid_field.get("inherit_above", False):
                        forward_fill_cols[tag_name].add(field_name)

                    known_cols[tag_name].add(field_name)

                    for valid_field_name in valid_field_names:
                        valid_column_names[tag_name][valid_field_name] = field_name

                    row_comment_chars[tag_name][field_name] = valid_field.get(
                        "row_ignore_symbol", list()
                    )

            # TODO: Account for differences in valid field names with base_tag
            if "base_tag" in tag_info:
                base_tag = to_tag(tag_info["base_tag"])
                mod_type = tag_info["mod_type"]
                if base_tag in valid_column_names:
                    valid_column_names[tag_name] = valid_column_names[base_tag]
                    discard_if_empty.append(tag_name)
                if base_tag in column_default_value:
                    column_default_value[tag_name] = column_default_value[base_tag]
                if base_tag in row_comment_chars:
                    row_comment_chars[tag_name] = row_comment_chars[base_tag]
                if base_tag in query_cols:
                    query_cols[tag_name] = query_cols[base_tag].difference({mod_type})
                if base_tag in lists_cols:
                    lists_cols[tag_name] = lists_cols[base_tag].difference({mod_type})
                if base_tag in known_cols:
                    known_cols[tag_name] = known_cols[base_tag].difference({mod_type})
                if base_tag in add_cols:
                    add_cols[tag_name] = add_cols[base_tag].difference({mod_type})
                if base_tag in required_cols:
                    required_cols[tag_name] = required_cols[base_tag].difference(
                        {mod_type}
                    )
                if base_tag in forward_fill_cols:
                    forward_fill_cols[tag_name] = forward_fill_cols[
                        base_tag
                    ].difference({mod_type})

        return (
            valid_column_names,
            column_default_value,
            row_comment_chars,
            discard_if_empty,
            query_cols,
            lists_cols,
            known_cols,
            required_cols,
            add_cols,
            forward_fill_cols,
        )

    @staticmethod
    def _read_veda_attr_defaults(
        veda_attr_defaults_file: str, attr_mappings: list[TimesXlMap]
    ) -> tuple[dict[str, dict[str, list]], set[str]]:
        # Read veda_tags_file
        with resources.open_text("xl2times.config", veda_attr_defaults_file) as f:
            defaults = json.load(f)

        veda_attr_defaults = {
            "aliases": defaultdict(list),
            "commodity": {},
            "other_indexes": {},
            "cg": {},
            "limtype": {"FX": [], "LO": [], "UP": []},
            "tslvl": {"DAYNITE": [], "ANNUAL": []},
            "year2": {"EOH": []},
        }

        group_defaults = {"limtype", "tslvl", "year2"}
        individual_defaults = {"commodity", "other_indexes", "cg"}

        attr_aliases = {
            attr for attr in defaults if "times-attribute" in defaults[attr]
        }

        for attr, attr_info in defaults.items():
            # Populate aliases by attribute dictionary
            if "times-attribute" in attr_info:
                times_attr = attr_info["times-attribute"]
                veda_attr_defaults["aliases"][times_attr].append(attr)

            if "defaults" in attr_info:
                attr_defaults = attr_info["defaults"]

                for item in group_defaults.intersection(attr_defaults.keys()):
                    group_value = attr_defaults[item]
                    veda_attr_defaults[item][group_value].append(attr)

                for item in individual_defaults.intersection(attr_defaults.keys()):
                    veda_attr_defaults[item][attr] = attr_defaults[item]

        # Specify default values for the attributes that are not defined in the file
        attr_with_cg = {
            attr_mapping.times_name
            for attr_mapping in attr_mappings
            if "cg" in set(attr_mapping.xl_cols)
        }
        for attr in attr_with_cg.difference(veda_attr_defaults["cg"].keys()):
            veda_attr_defaults["cg"][attr] = ["other_indexes"]

        return veda_attr_defaults, attr_aliases

    @staticmethod
    def _read_regions_filter(regions_list: str) -> set[str]:
        if regions_list == "":
            return set()
        else:
            return set(regions_list.strip(" ").upper().split(sep=","))



import re
import time

import numpy
from loguru import logger
from openpyxl import load_workbook
from openpyxl.worksheet.cell_range import CellRange
from pandas.core.frame import DataFrame

from . import datatypes, utils


def extract_tables(filename: str) -> list[datatypes.EmbeddedXlTable]:
    """Run the extract_table function on each individual table in each worksheet of the
    given excel file.

    Parameters
    ----------
    filename
        Path to the excel file we will extract tables from.

    Returns
    -------
    list[datatypes.EmbeddedXlTable]
        List of table objects in EmbeddedXlTable format.
    """
    start_time = time.time()

    workbook = load_workbook(filename=filename, data_only=True)

    tables = []
    for sheet in workbook.worksheets:
        # Creating dataframe with dtype=object solves problems with ints being cast to floats
        # https://stackoverflow.com/questions/40251948/stop-pandas-from-converting-int-to-float-due-to-an-insertion-in-another-column
        df = DataFrame(sheet.values, dtype=object)
        active_uc_sets = {}
        formatted_uc_sets = {}
        sheet_tables = []

        for row_index, row in df.iterrows():
            for colname in df.columns:
                value = str(row[colname])
                if value.startswith("~"):
                    match = re.match(
                        f"{datatypes.Tag.uc_sets.value}:(.*)", value, re.IGNORECASE
                    )
                    if match:
                        updated, active_uc_sets = _update_uc_sets(
                            active_uc_sets, match.group(1)
                        )
                        if updated:
                            formatted_uc_sets = {
                                k: v
                                for type, values in active_uc_sets.items()
                                for k, v in values.items()
                            }
                        else:
                            logger.warning(
                                f"Malformed {match.group(0)} in {sheet.title}, {filename}"
                            )
                    else:
                        col_index = df.columns.get_loc(colname)
                        sheet_tables.append(
                            extract_table(
                                row_index,
                                col_index,
                                formatted_uc_sets,
                                df,
                                sheet.title,
                                filename,
                            )
                        )

        tables += sheet_tables

    end_time = time.time()
    if end_time - start_time > 2:
        logger.info(f"Loaded {filename} in {end_time-start_time:.2f} seconds")

    return tables


def extract_table(
    tag_row: int,
    tag_col: int,
    uc_sets: dict[str, str],
    df: DataFrame,
    sheetname: str,
    filename: str,
) -> datatypes.EmbeddedXlTable:
    """For each individual table tag found in a worksheet, this function aims to extract
    the associated table.

    We recognise several types of tables:

    - Single cell tables: Tables with only one value, either below or to the right of
      the table tag. We interpret these as a single data item with
      a column name VALUE.
    - Multiple cell tables: Tables with multiple values, possibly extending accross
      several rows and columns. We delimitate them using empty
      spaces around them and the column names are determined by the
      values in the row immediately below the table tag

    Parameters
    ----------
    tag_row
        Row number for the tag designating the table to be extracted
    tag_col
        Column number for the tag designating the table to be extracted
    uc_sets
        Sets (regions and timeslices) for user constraints
    df
        Dataframe object containing all values for the worksheet being evaluated
    sheetname
        Name of the worksheet being evaluated
    filename
        Path to the excel file being evaluated.

    Returns
    -------
        Table object in the EmbeddedXlTable format.
    """
    # If the cell to the right is not empty then we read a scalar from it
    # Otherwise the row below is the header
    if df.shape[1] > tag_col + 1 and not cell_is_empty(df.iloc[tag_row, tag_col + 1]):
        table_range = str(
            CellRange(
                min_col=tag_col + 2,
                min_row=tag_row + 1,
                max_col=tag_col + 2,
                max_row=tag_row + 1,
            )
        )
        table_df = DataFrame(columns=["VALUE"])
        table_df.loc[0] = [df.iloc[tag_row, tag_col + 1]]
        uc_sets = {}
    else:
        header_row = tag_row + 1

        start_col = tag_col
        while start_col > 0 and not cell_is_empty(df.iloc[header_row, start_col - 1]):
            start_col -= 1

        end_col = tag_col
        while end_col < df.shape[1] and not cell_is_empty(df.iloc[header_row, end_col]):
            end_col += 1

        end_row = header_row
        while end_row < df.shape[0] and not are_cells_all_empty(
            df, end_row, start_col, end_col
        ):
            end_row += 1

        # Excel cell numbering starts at 1, while pandas starts at 0
        table_range = str(
            CellRange(
                min_col=start_col + 1,
                min_row=header_row + 1,
                max_col=end_col + 1,
                max_row=end_row + 1,
            )
        )

        if end_row - header_row == 1 and end_col - start_col == 1:
            # Interpret single cell tables as a single data item with a column name VALUE
            table_df = DataFrame(df.iloc[header_row, start_col:end_col])
            table_df.columns = ["VALUE"]
        else:
            table_df = df.iloc[header_row + 1 : end_row, start_col:end_col]
            # Make all columns names strings as some are integers e.g. years
            table_df.columns = [str(x) for x in df.iloc[header_row, start_col:end_col]]

    table_df = table_df.reset_index(drop=True)

    # Don't use applymap because it can convert ints to floats
    # https://pandas.pydata.org/pandas-docs/stable/user_guide/gotchas.html#gotchas-intna
    for i in range(table_df.shape[0]):
        for j in range(table_df.shape[1]):
            value = table_df.iat[i, j]
            if isinstance(value, float):
                table_df.iat[i, j] = utils.round_sig(value, 15)

    return datatypes.EmbeddedXlTable(
        filename=filename,
        sheetname=sheetname,
        range=table_range,
        tag=df.iloc[tag_row, tag_col],
        uc_sets=uc_sets,
        dataframe=table_df,
    )


def are_cells_all_empty(df, row: int, start_col: int, end_col: int) -> bool:
    """Check if all cells in a given row are empty by calling cell_is_empty() on them.

    Parameters
    ----------
    df
        Dataframe object containing all values for the worksheet being evaluated
    row
        Row of the dataframe to be evaluated.
    start_col
        Initial column of the dataframe to be evaluated.
    end_col
        Final column of the dataframe to be evaluated.

    Returns
    -------
    bool
        Boolean indicating if all the cells are empty.
    """
    for col in range(start_col, end_col):
        if not cell_is_empty(df.iloc[row, col]):
            return False
    return True


def cell_is_empty(value) -> bool:
    """Check if the given cell is empty.

    Parameters
    ----------
    value
        Cell value.

    Returns
    -------
    bool
        Boolean indicating if the cells are empty.
    """
    return (
        value is None
        or (isinstance(value, numpy.floating) and numpy.isnan(value))
        or (isinstance(value, str) and len(value.strip()) == 0)
    )


def _update_uc_sets(
    uc_sets: dict[str, dict[str, str]], new_element: str
) -> tuple[bool, dict[str, dict[str, str]]]:

    # Overview of the sets: https://times.readthedocs.io/en/latest/part-4/part-4.html#uc-sets-in-veda2
    # Categorise by type
    mapping = {
        "R_E": "region",
        "R_S": "region",
        "T_E": "period",
        "T_S": "period",
        "T_SUC": "period",
        "TS_E": "timeslice",
        "TS_S": "timeslice",
    }
    updated = False
    parts = [new_element.strip() for new_element in new_element.split(":")]
    # Check if the new element seems valid and if it is, update the uc_sets
    if len(parts) == 2 and parts[0] in mapping:
        uc_sets[mapping[parts[0]]] = {parts[0]: parts[1]}
        updated = True

    return updated, uc_sets



import argparse
import json
import os
import sys
from collections import defaultdict
from pathlib import Path

import numpy as np
import pandas as pd
from loguru import logger


def parse_parameter_values_from_file(
    path: Path,
) -> tuple[dict[str, list], dict[str, set]]:
    """Parse a `*.dd` file and extract the sets and parameters.

    There are parameters and sets, and each has a slightly different format.
    `*.dd` files have data of the following form:

    PARAMETER
    PARAM_NAME ' '/
    attr_1 attr_2
    ...

    /

    SET SET_NAME
    /
    attr_1.attr_2.attr_3
    ...

    /
    """
    data = list(open(path))
    data = [line.rstrip() for line in data]

    param_value_dict: dict[str, list] = dict()
    set_data_dict: dict[str, set] = dict()
    index = 0
    while index < len(data):
        if data[index].startswith("PARAMETER"):
            # We expect the parameter name on the next line.
            index += 1
            while data[index].strip() == "":
                index += 1

            param_name = data[index].replace(
                " ' '/", ""
            )  # param_name is followed by this pattern

            index += 1
            param_data = []
            # Parse values until a line with / is encountered.
            while not data[index].startswith("/") and data[index] != "":
                line = data[index]

                # Either "value" for a scalar, or "key value" for an array.
                # So value is always the last word, or only token
                split_point = line.rfind(" ")
                if split_point == -1:
                    # if only one word
                    attributes, value = [], line
                else:
                    attributes, value = line[:split_point], line[split_point + 1 :]
                    attributes = attributes.split(".")
                    attributes = [a if " " in a else a.strip("'") for a in attributes]

                param_data.append([*attributes, value])

                index += 1

            param_value_dict[param_name] = param_data

        if data[index].startswith("SET"):
            # See https://www.gams.com/latest/docs/UG_SetDefinition.html
            # This can only parse a subset of the allowed representations
            _, name = data[index].split(" ")
            index += 1
            while data[index].strip() == "":
                index += 1

            assert data[index].startswith("/")
            index += 1

            set_data = set()
            while not data[index].startswith("/") and data[index] != "":
                parts = [[]]
                for word in data[index].split("'"):
                    if word != "":
                        if word.isspace():
                            parts.append([])
                        elif word.endswith(" "):
                            parts[-1].append(word.strip())
                            parts.append([])
                        else:
                            parts[-1].append(word)
                words = ["".join(part) for part in parts]
                attributes = words[0].split(".")
                if len(words) == 1:
                    set_data.add(tuple([*attributes]))
                elif len(words) == 2:
                    text = words[1]
                    set_data.add(tuple([*attributes, text]))
                else:
                    raise ValueError(
                        f"Unexpected number of spaces in set value setting: {data[index]}"
                    )

                index += 1

            if name in set_data_dict:
                set_data_dict[name].update(set_data)
            else:
                set_data_dict[name] = set_data

        index += 1

    return param_value_dict, set_data_dict


def save_data_with_headers(
    param_data_dict: dict[str, pd.DataFrame | list[str]],
    headers_data: dict[str, list[str]],
    save_dir: str,
) -> None:
    """Saves data (with headers) to the provided directory.

    Note that the header and data dictionaries are assumed to be parallel dictionaries

    Parameters
    ----------
    param_data_dict
        Dictionary containing key=param_name and val=dataframe for parameters or List[str] for sets
    headers_data
        Dictionary containing key=param_name and val=dataframes
    save_dir
        Path to folder in which to save the tabular data files
    """
    for param_name, param_data in param_data_dict.items():
        try:
            columns = headers_data[param_name]
        except KeyError:
            raise ValueError(
                f"Could not find mapping for {param_name} in mapping file."
            )
        for row in param_data:
            if len(row) != len(columns):
                raise ValueError(
                    f"Mismatched number of columns for param {param_name} between data ({len(row)}) and mapping ({len(columns)})"
                )
        df = pd.DataFrame(
            data=np.asarray(param_data)[:, 0 : len(columns)], columns=columns
        )
        df.to_csv(os.path.join(save_dir, param_name + ".csv"), index=False)

    return


def generate_headers_by_attr() -> dict[str, list[str]]:
    with open("xl2times/config/times-info.json") as f:
        attributes = json.load(f)

    headers_by_attr = {}

    for attr in attributes:
        if attr["gams-cat"] == "parameter":
            headers_by_attr[attr["name"]] = [*attr["indexes"], "VALUE"]
        else:
            headers_by_attr[attr["name"]] = attr["indexes"]

    return headers_by_attr


def convert_dd_to_tabular(
    dd_files: list[Path], output_dir: str, headers_by_attr: dict[str, list[str]]
) -> None:

    all_sets = defaultdict(list)
    all_parameters = defaultdict(list)
    for path in dd_files:
        logger.info(f"Processing path: {path}")
        local_param_values, local_sets = parse_parameter_values_from_file(path)

        # merge params from file into global collection
        for param, data in local_param_values.items():
            all_parameters[param].extend(data)

        for set_name, data in local_sets.items():
            all_sets[set_name].extend(data)

    use_subfolders = False
    if use_subfolders:
        param_path = os.path.join(output_dir, "params")
        set_path = os.path.join(output_dir, "sets")
    else:
        param_path = output_dir
        set_path = output_dir
    os.makedirs(param_path, exist_ok=True)
    os.makedirs(set_path, exist_ok=True)

    # Extract headers with key=param_name and value=List[attributes]
    lines = list(open("xl2times/config/times_mapping.txt"))
    headers_data = headers_by_attr
    # The following will overwrite data obtained from headers_by_attr
    # TODO: Remove once migration is done?
    for line in lines:
        ln = line.strip()
        if ln != "":
            param_name = ln.split("[")[0]
            attributes = ln.split("[")[1].split("]")[0].split(",")
            headers_data[param_name] = [*attributes]

    save_data_with_headers(all_parameters, headers_data, param_path)
    save_data_with_headers(all_sets, headers_data, set_path)

    return


def main(arg_list: None | list[str] = None):
    args_parser = argparse.ArgumentParser()
    args_parser.add_argument(
        "input_dir", type=str, help="Input directory containing .dd files."
    )
    args_parser.add_argument(
        "output_dir", type=str, help="Output directory to save the .csv files in."
    )
    args_parser.add_argument(
        "--include_files",
        type=str,
        nargs="*",
        help="List of .dd file stems to process.",
    )
    args = args_parser.parse_args(arg_list)
    dd_files = [p for p in Path(args.input_dir).glob("*.dd")]
    if args.include_files:
        valid_stems = {stem.lower() for stem in args.include_files}
        dd_files = [p for p in dd_files if p.stem.lower() in valid_stems]
    convert_dd_to_tabular(dd_files, args.output_dir, generate_headers_by_attr())


if __name__ == "__main__":
    main(sys.argv[1:])



import pandas as pd
from pandas import DataFrame

from xl2times import transforms, utils
from xl2times.transforms import (
    _count_comm_group_vectorised,
    _match_wildcards,
    _process_comm_groups_vectorised,
    commodity_map,
    process_map,
)

utils.setup_logger(None)

pd.set_option("display.max_rows", 20)
pd.set_option("display.max_columns", 20)
pd.set_option("display.width", 300)
pd.set_option("display.max_colwidth", 75)
pd.set_option("display.precision", 3)


class TestTransforms:
    def test_explode_process_commodity_cols(self):
        df = DataFrame(
            {
                "process": ["a", "b", ["c", "d"]],
                "commodity": [["v", "w", "x"], "y", "z"],
            }
        )
        df2 = transforms.explode_process_commodity_cols(
            None,  # pyright: ignore
            {"name": df.copy()},
            None,  # pyright: ignore
        )
        correct = DataFrame(
            {
                "process": ["a", "a", "a", "b", "c", "d"],
                "commodity": ["v", "w", "x", "y", "z", "z"],
            }
        )
        assert df2["name"].equals(correct)

    def test_uc_wildcards(self):
        """Tests logic that matches wildcards in the process_uc_wildcards transform."""
        import pickle

        df_in = pd.read_parquet("tests/data/process_uc_wildcards_ireland_data.parquet")
        with open("tests/data/process_uc_wildcards_ireland_dict.pkl", "rb") as f:
            dictionary = pickle.load(f)
        df = df_in.copy()

        for result_col, item_map in {
            "process": process_map,
            "commodity": commodity_map,
        }.items():
            df = _match_wildcards(df, item_map, dictionary, result_col)

        # unit tests
        assert df is not None and not df.empty
        assert (
            df.shape[0] >= df_in.shape[0]
        ), "should have more rows after processing uc_wildcards"
        assert (
            df.shape[1] < df_in.shape[1]
        ), "should have fewer columns after processing uc_wildcards"
        assert "process" in df.columns, "should have added process column"
        assert "commodity" in df.columns, "should have added commodity column"

    def test_generate_commodity_groups(self):
        """Tests that the _count_comm_group_vectorised function works as expected.

        Full austimes run:
            Vectorised version took 0.021999 seconds
            looped version took 966.653371 seconds
            43958x speedup
        """
        # data extracted immediately before the original for loops
        comm_groups = pd.read_parquet(
            "tests/data/comm_groups_austimes_test_data.parquet"
        ).drop(columns=["commoditygroup"])

        # filter data so test runs faster
        comm_groups = comm_groups.query("region in ['ACT', 'NSW']")

        comm_groups2 = comm_groups.copy()
        _count_comm_group_vectorised(comm_groups2)
        assert comm_groups2.drop(columns=["commoditygroup"]).equals(comm_groups)
        assert comm_groups2.shape == (comm_groups.shape[0], comm_groups.shape[1] + 1)

    def test_default_pcg_vectorised(self):
        """Tests the default primary commodity group identification logic runs
        correctly.

        Full austimes run:
            Looped version took 1107.66 seconds
            Vectorised version took 62.85 seconds
        """
        # data extracted immediately before the original for loops
        comm_groups = pd.read_parquet("tests/data/austimes_pcg_test_data.parquet")

        comm_groups = comm_groups[(comm_groups["region"].isin(["ACT", "NT"]))]
        comm_groups2 = _process_comm_groups_vectorised(
            comm_groups.copy(), transforms.csets_ordered_for_pcg
        )
        assert comm_groups2 is not None and not comm_groups2.empty
        assert comm_groups2.shape == (comm_groups.shape[0], comm_groups.shape[1] + 1)
        assert comm_groups2.drop(columns=["DefaultVedaPCG"]).equals(comm_groups)


if __name__ == "__main__":
    # TestTransforms().test_default_pcg_vectorised()
    TestTransforms().test_uc_wildcards()
